import { Component, OnInit, Input, OnChanges, ContentChild, TemplateRef, Output, EventEmitter, ViewChild, SimpleChanges, ContentChildren, NgZone } from '@angular/core';
import { HttpService } from 'src/app/core/http.service';
import { MatTableDataSource } from '@angular/material/table';
import { MatPaginator } from '@angular/material/paginator';
import { MatSort, Sort } from '@angular/material/sort';
import { CdkDragDrop, moveItemInArray } from '@angular/cdk/drag-drop';
import { Globals } from 'src/app/core/global';
import { filter, take } from 'rxjs/operators';
import { Subject } from 'rxjs';

/**
 * How to use the table?
 * you can just add element into yout html
 * @param {string} tableName Field TABLE_NAME.
 * @param {Array<any>} tableData Detail row list in table.
 * @param {Array<any>} tableHeader
 * @param {Array<any>} tableFooter show footer
 * @param {boolean} showPagination
 * @param {boolean} sortTable
 * @param {string} stickyCol
 * @param {string} stickyEndCol
 * @event editClick($event) Send function name to this for Edit button action.
 * 
 * @tableFooter param list
 *  1. fieldName: show under the header fieldName
 *  2. show: true | false (show column in footer)
 *  3. value: fix value of the field (If not defined, table will sum field value)
 * 
 * you can custom condition to show another value inside the column
 * @var {object} header Current header row
 * @var {object} data Current data row
 * @description<ng-template let-header="header" let-data="data">
 * @description--> insert your condition
 * @description</ng-template>
 */

@Component({
  selector: 'auto-table',
  templateUrl: './auto-table.component.html',
  styleUrls: ['./auto-table.component.scss']
})
export class AutoTableComponent implements OnChanges {
  public headerList: Array<any> = [];
  public dataList: any;
  public displayedColumns: string[] = [];
  public displayFooter: Array<any> = [];
  public isCustomContent: boolean = false;
  public isCustomHeader: boolean = false;
  public templateBody: TemplateRef<any>;
  public templateHeader: TemplateRef<any>;

  public progressBar: any;
  public showProgressBar: boolean = false;
  public showNoRecord: boolean = false;
  private dataTypes: Array<string> = [];
  public _filter: any = {};

  public i: number;
  private tableFooterChanged: Subject<Array<any>> = new Subject<Array<any>>();

  @Input("tableName") tableName: string;
  @Input("tableData") tableData: Array<any> = [];
  @Input("tableHeader") tableHeader: Array<any> = [];
  @Input("tableFooter") tableFooter: Array<any> = [];
  @Input("showPagination") showPagination: boolean = true;
  @Input("sortTable") sortTable: boolean = true;
  @Input("maxHeight") maxHeight: string = "82vh";
  @Input("stickyCol") stickyCol: string | Array<string>;
  @Input("stickyEndCol") stickyEndCol: string | Array<string>;
  @Input("hideColumns") hideColumns: Array<string> = [];
  @Input("minpage") minpage: number = 10;
  @Input("editIcon") editIcon: string = "edit";
  @Output("editClick") editClick = new EventEmitter<any>();
  @Output("filteredData") filteredData = new EventEmitter<any>();
  @Input("themeColor") themeColor: string = '1';

  @ContentChildren(TemplateRef) templateRef;
  @ViewChild("table") table: any;
  @ViewChild(MatPaginator, { static: true }) paginator: MatPaginator;
  @ViewChild(MatSort, { static: true }) sort: MatSort;

  constructor(
    private http: HttpService,
    public global: Globals,
    private ngZone: NgZone
  ) { }

  ngOnInit(): void {
    if (this.tableName) {
      this.getTableHeader();
    } else if (this.tableHeader.length > 0) {
      this.headerList = this.tableHeader;
      this.ArrangeShowColumn(this.headerList);
    }

    this.tableFooterChanged.pipe(
      filter(v => v && v.length > 0)
    ).subscribe(v => {
      this.setFooterItems(v)
    })
  }

  ngOnChanges(changes: SimpleChanges): void {
    if (!changes["tableData"].isFirstChange() && this.tableData) {
      if (this.tableData.length === 0) {
        if (!this.showProgressBar) {
          this.loading();
        } else {
          this.unloading();
          this.showNoRecord = true;
        }
      } else {
        this.unloading();
      }
    }

    this.dataList = new MatTableDataSource<any>(this.tableData || []);
    if (this.showPagination) {
      this.dataList.paginator = this.paginator;
    }
    this.dataList.sort = this.sort;

    this.tableFooterChanged.next(this.tableFooter);

    this.dataList.filterPredicate = (data, filterObject: string): boolean => {
      const filters: any = JSON.parse(filterObject);

      return Object.keys(filters)
        .map(column => (data[column] || "").toString().includes(filters[column].toString()))
        .reduce((acc: boolean, curr: boolean) => (acc == curr) && acc, true);
    };
    this.dataList.filter = JSON.stringify(this._filter);
    this.filteredData.emit(this.dataList.filteredData);

    this.ArrangeShowColumn(this.headerList);

  }

  ngAfterContentInit(): void {
    this.isCustomContent = false;
    this.isCustomHeader = false;
    if (this.templateRef) {
      if (this.templateRef.length == 1) {
        this.isCustomContent = true;
        this.templateBody = this.templateRef._results[0];
      } else if (this.templateRef.length == 2
        && this.templateRef._results.find(t => t._declarationTContainer?.localNames?.includes("tblHeader"))
        && this.templateRef._results.find(t => t._declarationTContainer?.localNames?.includes("tblBody"))) {
        let headerIndex = this.templateRef._results.findIndex(t => t._declarationTContainer?.localNames?.includes("tblHeader"));
        let bodyIndex = this.templateRef._results.findIndex(t => t._declarationTContainer?.localNames?.includes("tblBody"));
        this.isCustomContent = true;
        this.isCustomHeader = true;
        this.templateHeader = this.templateRef._results[headerIndex];
        this.templateBody = this.templateRef._results[bodyIndex];
      }
    }
  }

  ngAfterViewChecked() {
    this.ngZone.onMicrotaskEmpty.pipe(take(3)).subscribe(() => this.table.updateStickyColumnStyles());
  }

  getTableHeader() {
      this.http.get("System", "GetHeaderGridView", this.tableName)
      .subscribe(result => {
        this.headerList = result.filter(v => v.VISIBILITY === true);
        
        //Change footer for header column map
        this.tableFooterChanged.next(this.tableFooter);

        this.ArrangeShowColumn(this.headerList);
      })
  }

  setFooterItems(footers: Array<any>) {

    this.displayFooter = [];

    this.displayFooter = this.headerList.map(h => {
      let fieldName = h.FIELD_NAME;
      let value, colspan;
      let footer = footers.find(f => f.fieldName === fieldName);
      if (footer && footer.show) {
        if ((footer.value === null || footer.value === undefined) && (h.DATA_TYPE === 'number' || h.DATA_TYPE === 'number2' || h.DATA_TYPE === 'number3' || h.DATA_TYPE === 'number5')) {
          let sumValue = this.tableData.reduce((v, o) => parseFloat(o[fieldName]) === o[fieldName] ? v + parseFloat(o[fieldName]) : v, 0);

          value = sumValue || 0;
        } else if (footer.value) {
          value = footer.value;
        } else {
          value = 0;
        }

        colspan = footer.colspan;
      }

      return { FIELD_NAME: fieldName, VALUE: value, COLSPAN: colspan || 1 };

    })
  }

  getFooterValue(fieldName) {
    //console.log('this.displayFooter',this.displayFooter)
    let findFooterItem = this.displayFooter.find(v => v.FIELD_NAME === fieldName);
    return findFooterItem ? findFooterItem.VALUE : null;
  }

  getFooterColspan(fieldName) {
    let findFooterItem = this.displayFooter.find(v => v.FIELD_NAME === fieldName);
    return findFooterItem ? findFooterItem.COLSPAN : null;
  }


  editBtnClick(item: any) {
    this.editClick.emit(item);
  }

  loading(isFirstCall: boolean = false) {
    this.showProgressBar = true;
    this.showNoRecord = false;

    clearTimeout(this.progressBar);

    this.progressBar = setTimeout(() => {
      this.showProgressBar = false;

      if (!isFirstCall && this.tableData.length === 0) {
        this.showNoRecord = true;
      }
    }, 5000);
  }

  unloading() {
    clearTimeout(this.progressBar);
    this.showProgressBar = false;
    this.showNoRecord = false;
  }

  applyFilter(column, isClear: boolean) {
    if (isClear || this._filter[column] === '') delete this._filter[column];

    this.dataList.filter = JSON.stringify(this._filter);
    this.filteredData.emit(this.dataList.filteredData);
    
  }

  sortUp(column: string) {
    this.sort.active = column;
    // this.dataList.sort = this.sort;
    const sortState: Sort = {
      active: column,
      direction: 'asc'
    };

    this.sort.direction = sortState.direction;
    this.sort.sortChange.emit(sortState);
  }

  sortDown(column: string) {
    this.sort.active = column;
    const sortState: Sort = {
      active: column,
      direction: 'desc'
    };

    this.sort.direction = sortState.direction;
    this.sort.sortChange.emit(sortState);
  }

  checkSticky(cols: string | Array<string>, fieldName: string) {

    if (cols) {
      if (typeof cols === 'string') {
        return cols === fieldName;
      } else {
        return cols.indexOf(fieldName) >= 0;
      }
    }

    return false;
  }

  ArrangeShowColumn(data) {
    this.displayedColumns = data.filter(v => this.hideColumns.indexOf(v.FIELD_NAME) < 0)
      .map(v => v.FIELD_NAME);
  }

  public CheckIsFinite(value: number) {
    return Number.isFinite(value);
  }
}
